# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from abc import ABC

from ..models import Role, RoleType
from ..utils import load_json


class CoreGetters(ABC):
    @property
    def presets(self) -> dict[str, dict]:
        if not self._presets:
            self._presets = load_json(self._file_presets)
        return self._presets

    @property
    def roles(self) -> dict[RoleType, Role]:
        if not self._roles:
            roles = load_json(self._file_roles)
            self._roles = {
                RoleType(key): Role(role_type=RoleType(key), **value)
                for key, value in roles.items()
            }
        return self._roles

    def role(self, role_type: RoleType) -> Role | None:
        return self.roles.get(role_type, None)

    @property
    def flash(self) -> dict[str, str]:
        if not self._flash:
            self._flash = load_json(self._file_flash)
        return self._flash



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from svgwrite import Drawing, shapes

from .base import Shape
from .fill_style import FillStyle


class Circle(Shape):
    r: float = None
    d: float = None
    fill: FillStyle = None
    stroke: FillStyle = None

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        if not self.r and not self.d:
            raise ValueError("No radius or diameter")
        if self.d:
            self.r = self.d / 2

    def draw(self, dwg: Drawing):
        circle = shapes.Circle(
            center=self.pos.tuple,
            r=self.r,
            id=self.fullid,
        )
        if self.fill:
            self.fill.apply_to(dwg, circle)
        if self.stroke:
            self.stroke.apply_to(dwg, circle, stroke=True)
        dwg.add(circle)

    @property
    def x1(self) -> float:
        return self.pos.x - self.r

    @property
    def y1(self) -> float:
        return self.pos.y - self.r

    @property
    def x2(self) -> float:
        return self.pos.x + self.r

    @property
    def y2(self) -> float:
        return self.pos.y + self.r



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from typing import Any

from pydantic import BaseModel, Field


class HasVars(BaseModel):
    vars: dict = {}

    def var(self, name: str) -> str | None:
        return self.vars.get(name, None)


class HasId(BaseModel):
    id: str = Field(None, alias="id")
    name: str = Field(None, alias="name")

    id_suffix: str = None

    @property
    def fullid(self) -> str | None:
        suffix = "." + self.id_suffix if self.id_suffix else ""
        if self.id:
            return self.id + suffix
        if self.name:
            return self.name + suffix
        if isinstance(self, HasVars) and "NAME" in self.vars:
            return self.var("NAME") + suffix
        return None


ParentType = HasVars | HasId | Any



# Copyright (c) Kuba Szczodrzyński 2022-05-12.

from ..utils import Model, sizeof


class FlashRegion(Model):
    name: str
    offset: int
    length: int

    title: str = None
    read_only: bool = False
    hex_offs_len: str = 6
    hex_size_len: str = 0

    @property
    def start(self) -> int:
        return self.offset

    @property
    def end(self) -> int:
        return self.offset + self.length

    @property
    def start_hex(self) -> str:
        if not self.hex_offs_len:
            return "0x%X" % self.start
        return f"0x%0{self.hex_offs_len}X" % self.start

    @property
    def end_hex(self) -> str:
        if not self.hex_offs_len:
            return "0x%X" % self.end
        return f"0x%0{self.hex_offs_len}X" % self.end

    @property
    def length_hex(self) -> str:
        if not self.hex_size_len:
            return "0x%X" % self.length
        return f"0x%0{self.hex_size_len}X" % self.length

    @property
    def length_str(self) -> str:
        return sizeof(self.length)

    @property
    def lst(self) -> list[str]:
        return [self.start_hex, self.length_str + " / " + self.length_hex, self.end_hex]



# Copyright (c) Kuba Szczodrzyński 2022-06-15.

import os
from os.path import dirname

from ..core import Core
from ..models.board import Board
from ..models.enums import RoleType
from .features import PinFeatures
from .parts import VariantParts
from .section import SectionType


class VariantWriter(VariantParts):
    core: Core

    def __init__(self, core: Core) -> None:
        self.core = core
        self.pins = {}
        self.sections = {}
        self.sorted_pins = []
        self.sorted_sections = []

    def generate(self, board: Board):
        pcb = board.pcb
        if not pcb or not pcb.pinout:
            return
        if not [True for f in board.frameworks if "arduino" in f]:
            return

        interfaces: dict[RoleType, dict[int, dict[str, list[str]]]] = {
            RoleType.SPI: {},
            RoleType.I2C: {},
            RoleType.UART: {},
        }
        interface_ports: dict[RoleType, list[str]] = {
            RoleType.SPI: ["SCK", "MISO", "MOSI"],
            RoleType.I2C: ["SDA", "SCL"],
            RoleType.UART: ["RX", "TX"],
        }
        interface_sections: dict[RoleType, SectionType] = {
            RoleType.SPI: SectionType.SPI,
            RoleType.I2C: SectionType.WIRE,
            RoleType.UART: SectionType.SERIAL,
        }

        analog_alias = []

        for pin in pcb.pinout.values():
            pin_digital = pin.get(RoleType.ARD_D, None)
            pin_analog = pin.get(RoleType.ARD_A, None)
            c_name = pin.get(RoleType.C_NAME, None)
            gpio = pin.get(RoleType.GPIO, None)
            adc = pin.get(RoleType.ADC, None)

            comment = []
            hidden = ["ARD_A", "ARD_D", "IO", "C_NAME"]
            for role_type, roles in pin.items():
                if role_type.name in hidden:
                    continue
                role = self.core.role(role_type)
                if role:
                    comment += role.format(roles, long=True, hidden=hidden)
            comment = ", ".join(comment)

            if pin_digital:
                pin_name = pin_digital
                added = self.add_pin(pin_name, c_name or gpio, comment)
            elif pin_analog:
                pin_name = pin_analog
                added = self.add_pin(pin_name, c_name or str(adc), comment)
            else:
                continue

            if added:
                self.increment_item(SectionType.PINS, "PINS_COUNT")

            if pin_digital:
                self.add_pin_feature(pin_name, PinFeatures.PIN_GPIO)
                if added:
                    self.increment_item(SectionType.PINS, "NUM_DIGITAL_PINS")
            if pin_analog:
                self.add_pin_feature(pin_name, PinFeatures.PIN_ADC)
                if added:
                    self.increment_item(SectionType.PINS, "NUM_ANALOG_INPUTS")
                self.add_item(SectionType.ANALOG, f"PIN_{pin_analog}", pin_name)
                analog_alias.append(pin_analog)

            for role_type, roles in pin.items():
                if role_type not in interfaces:
                    continue
                if not isinstance(roles, list):
                    roles = [str(roles)]
                for role in roles:
                    # require "1_RX" format
                    if not role[0].isnumeric() or role[1] != "_":
                        continue
                    (index, _, port) = role.partition("_")
                    if index not in interfaces[role_type]:
                        interfaces[role_type][index] = {}
                    if port not in interfaces[role_type][index]:
                        interfaces[role_type][index][port] = []
                    interfaces[role_type][index][port].append(pin_name)

            if RoleType.PWM in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_PWM)
            if RoleType.I2C in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_I2C)
            if RoleType.I2S in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_I2S)
            if RoleType.IRQ in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_IRQ)
            if RoleType.JTAG in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_JTAG)
            if RoleType.SPI in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_SPI)
            if RoleType.SWD in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_SWD)
            if RoleType.UART in pin:
                self.add_pin_feature(pin_name, PinFeatures.PIN_UART)

        self.add_item(SectionType.PINS, "NUM_ANALOG_OUTPUTS", 0)

        for alias in analog_alias:
            self.add_item(SectionType.ANALOG, alias, f"PIN_{alias}")

        for role_type, intfs in interfaces.items():
            section = interface_sections[role_type]
            count = 0
            items = {}
            for idx in sorted(intfs.keys()):
                ports = intfs[idx]
                if not all(port in ports for port in interface_ports[role_type]):
                    continue
                count += 1
                for port in sorted(ports.keys()):
                    pins = ports[port]
                    key = f"PIN_{section.name}{idx}_{port}"
                    if len(pins) > 1:
                        for i, pin in enumerate(pins):
                            items[f"{key}_{i}"] = pin
                    else:
                        items[key] = pins[0]
            self.add_item(section, f"{section.name}_INTERFACES_COUNT", count)
            for key, value in items.items():
                self.add_item(section, key, value)

        self.prepare_data()

    def prepare_data(self):
        pins_d = []
        pins_a = []
        pins_idx = {}
        for name in sorted(self.pins.keys(), key=lambda x: int(x[1:])):
            (gpio, features, comment) = self.pins[name]
            if name[0] == "D":
                pins_d.append((name, gpio, features, comment))
            if name[0] == "A":
                pins_a.append((name, gpio, features, comment))
        self.sorted_pins = pins_d + pins_a
        for i, (name, gpio, _, _) in enumerate(self.sorted_pins):
            pins_idx[name] = (i, gpio)

        self.sorted_sections = []
        for type in SectionType:
            if type not in self.sections:
                continue
            section = self.sections[type]
            for i, (key, value, _) in enumerate(section):
                if not key.startswith("PIN_"):
                    continue
                (idx, gpio) = pins_idx[value]
                section[i] = (key, f"{idx}u", gpio)
            self.sorted_sections.append((type, self.sections[type]))

    def save_compat(self, output: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            f.write('#include "variant.h"\n')

    def save_h(self, output: str, board_name: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            lines = [
                f"/* This file was auto-generated from {board_name} using boardgen */",
                "",
                "#pragma once",
                "",
                "#include <WVariant.h>",
                "",
                self.format_sections(),
            ]
            f.write("\n".join(lines))

    def save_cpp(self, output: str, board_name: str):
        os.makedirs(dirname(output), exist_ok=True)
        with open(output, "w", encoding="utf-8") as f:
            lines = [
                f"/* This file was auto-generated from {board_name} using boardgen */",
                "",
                "#include <Arduino.h>",
                "",
                'extern "C" {',
                "",
                self.format_pins(),
                "",
                '} // extern "C"',
                "",
            ]
            f.write("\n".join(lines))



# Copyright (c) Kuba Szczodrzyński 2022-05-11.

from pydantic.color import Color

from ..utils import Model
from .enums import RoleType, RoleValue

roles_idxswap = [
    RoleType.UART,
    RoleType.SPI,
    RoleType.I2C,
    RoleType.I2S,
    RoleType.TMR,
]
roles_nameidx = [
    RoleType.ADC,
    RoleType.PWM,
    RoleType.WAKE,
    RoleType.SD,
]
roles_arduino = [
    RoleType.ARD_D,
    RoleType.ARD_A,
]


class Role(Model):
    role_type: RoleType
    title: str
    color: Color
    ratio: float = 1.8

    def format(
        self, functions: RoleValue, long: bool = False, hidden: list[str] = []
    ) -> list[str]:
        # force a list
        if not isinstance(functions, list):
            if not functions:
                functions = [self.role_type.name]
            else:
                functions = [str(functions)]

        out = []
        for function in functions:
            # swap number and function
            if self.role_type in roles_idxswap and function[0].isnumeric():
                # skip hidden texts
                if function[2:] in hidden:
                    continue
                if long:
                    function = self.role_type.name + function
                else:
                    function = function[2:] + function[0]
            # append role name to number
            elif self.role_type in roles_nameidx and function.isnumeric():
                function = self.role_type.name + function
            elif self.role_type in roles_nameidx and long:
                function = self.role_type.name + "_" + function
            # SWDIO / SWCLK
            elif self.role_type == RoleType.SWD:
                function = "SW" + function
            # convert voltage
            elif (
                self.role_type == RoleType.PWR and function.replace(".", "").isnumeric()
            ):
                function = float(function)
                # decimal voltage
                if int(function) == function:
                    function = f"{int(function)}V"
                # floating point voltage
                else:
                    function = str(function).replace(".", "V")
            # remove hidden roles
            if self.role_type.name in hidden or function in hidden:
                continue
            out.append(function)
        return out
